@@ -24,6 +24,8 @@
 
 #define GNOME_DESKTOP_USE_UNSTABLE_API
 
+#include <config.h>
+#include <glib/gi18n-lib.h>
 #include "libgnomeui/gnome-rr.h"
 #include <string.h>
 #include <X11/Xlib.h>
@@ -120,14 +122,16 @@ struct GnomeRRMode
 static GnomeRRCrtc *  crtc_new          (ScreenInfo         *info,
 					 RRCrtc              id);
 static void           crtc_free         (GnomeRRCrtc        *crtc);
-static void           crtc_initialize   (GnomeRRCrtc        *crtc,
-					 XRRScreenResources *res);
+static gboolean       crtc_initialize   (GnomeRRCrtc        *crtc,
+					 XRRScreenResources *res,
+					 GError            **error);
 
 /* GnomeRROutput */
 static GnomeRROutput *output_new        (ScreenInfo         *info,
 					 RROutput            id);
-static void           output_initialize (GnomeRROutput      *output,
-					 XRRScreenResources *res);
+static gboolean       output_initialize (GnomeRROutput      *output,
+					 XRRScreenResources *res,
+					 GError            **error);
 static void           output_free       (GnomeRROutput      *output);
 
 /* GnomeRRMode */
@@ -138,6 +142,22 @@ static void           mode_initialize   (GnomeRRMode        *mode,
 static void           mode_free         (GnomeRRMode        *mode);
 
 
+/* Errors */
+
+/**
+ * gnome_rr_error_quark:
+ *
+ * Returns the #GQuark that will be used for #GError values returned by the
+ * GnomeRR API.
+ *
+ * Return value: a #GQuark used to identify errors coming from the GnomeRR API.
+ */
+GQuark
+gnome_rr_error_quark (void)
+{
+    return g_quark_from_static_string ("gnome-rr-error-quark");
+}
+
 /* Screen */
 static GnomeRROutput *
 gnome_rr_output_by_id (ScreenInfo *info, RROutput id)
@@ -306,7 +326,8 @@ gather_clone_modes (ScreenInfo *info)
 static gboolean
 fill_out_screen_info (Display *xdisplay,
 		      Window xroot,
-		      ScreenInfo *info)
+		      ScreenInfo *info,
+		      GError **error)
 {
     XRRScreenResources *resources;
     
@@ -321,14 +342,18 @@ fill_out_screen_info (Display *xdisplay,
                                 &(info->max_width),
                                 &(info->max_height))) {
         /* XRR caught an error */
-        return False;
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+		     _("could not get the range of screen sizes"));
+        return FALSE;
     }
     
     gdk_flush ();
     if (gdk_error_trap_pop ())
     {
         /* Unhandled X Error was generated */
-        return False;
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_UNKNOWN,
+		     _("unhandled X error while getting the range of screen sizes"));
+        return FALSE;
     }
     
 #if 0
@@ -337,7 +362,11 @@ fill_out_screen_info (Display *xdisplay,
 	     screen->min_height, screen->max_height);
 #endif
     
+#if (RANDR_MAJOR == 1 && RANDR_MINOR >= 3)
+    resources = XRRGetScreenResourcesCurrent (xdisplay, xroot);
+#else
     resources = XRRGetScreenResources (xdisplay, xroot);
+#endif
     
     if (resources)
     {
@@ -388,10 +417,16 @@ fill_out_screen_info (Display *xdisplay,
 	
 	/* Initialize */
 	for (crtc = info->crtcs; *crtc; ++crtc)
-	    crtc_initialize (*crtc, resources);
+	{
+	    if (!crtc_initialize (*crtc, resources, error))
+		return FALSE;
+	}
 	
 	for (output = info->outputs; *output; ++output)
-	    output_initialize (*output, resources);
+	{
+	    if (!output_initialize (*output, resources, error))
+		return FALSE;
+	}
 	
 	for (i = 0; i < resources->nmode; ++i)
 	{
@@ -409,13 +444,14 @@ fill_out_screen_info (Display *xdisplay,
 #if 0
 	g_print ("Couldn't get screen resources\n");
 #endif
-	
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+		     _("could not get the screen resources (CRTCs, outputs, modes)"));
 	return FALSE;
     }
 }
 
 static ScreenInfo *
-screen_info_new (GnomeRRScreen *screen)
+screen_info_new (GnomeRRScreen *screen, GError **error)
 {
     ScreenInfo *info = g_new0 (ScreenInfo, 1);
     
@@ -426,36 +462,36 @@ screen_info_new (GnomeRRScreen *screen)
     info->modes = NULL;
     info->screen = screen;
     
-    if (fill_out_screen_info (screen->xdisplay, screen->xroot, info))
+    if (fill_out_screen_info (screen->xdisplay, screen->xroot, info, error))
     {
 	return info;
     }
     else
     {
-	g_free (info);
+	screen_info_free (info);
 	return NULL;
     }
 }
 
 static gboolean
-screen_update (GnomeRRScreen *screen, gboolean force_callback)
+screen_update (GnomeRRScreen *screen, gboolean force_callback, GError **error)
 {
     ScreenInfo *info;
     gboolean changed = FALSE;
     
     g_assert (screen != NULL);
-    
-    info = screen_info_new (screen);
-    if (info)
-    {
-	if (info->resources->configTimestamp != screen->info->resources->configTimestamp)
+
+    info = screen_info_new (screen, error);
+    if (!info)
+	    return FALSE;
+
+    if (info->resources->configTimestamp != screen->info->resources->configTimestamp)
 	    changed = TRUE;
 	
-	screen_info_free (screen->info);
+    screen_info_free (screen->info);
 	
-	screen->info = info;
-    }
-    
+    screen->info = info;
+
     if ((changed || force_callback) && screen->callback)
 	screen->callback (screen, screen->data);
     
@@ -483,7 +519,7 @@ screen_on_event (GdkXEvent *xevent,
 	/* FIXME: we may need to be more discriminating in
 	 * what causes 'changed' events
 	 */
-	screen_update (screen, TRUE);
+	screen_update (screen, TRUE, NULL); /* NULL-GError */
     }
     
     /* Pass the event on to GTK+ */
@@ -496,11 +532,14 @@ screen_on_event (GdkXEvent *xevent,
 GnomeRRScreen *
 gnome_rr_screen_new (GdkScreen *gdk_screen,
 		     GnomeRRScreenChanged callback,
-		     gpointer data)
+		     gpointer data,
+		     GError **error)
 {
     Display *dpy = GDK_SCREEN_XDISPLAY (gdk_screen);
     int event_base;
     int ignore;
+
+    g_return_val_if_fail (error == NULL || *error == NULL, NULL);
     
     if (XRRQueryExtension (dpy, &event_base, &ignore))
     {
@@ -517,7 +556,7 @@ gnome_rr_screen_new (GdkScreen *gdk_screen,
 	
 	screen->randr_event_base = event_base;
 	
-	screen->info = screen_info_new (screen);
+	screen->info = screen_info_new (screen, error);
 	
 	if (!screen->info) {
 	    g_free (screen);
@@ -538,8 +577,13 @@ gnome_rr_screen_new (GdkScreen *gdk_screen,
 	gdk_window_add_filter (screen->gdk_root, screen_on_event, screen);
 	return screen;
     }
-    
-    return NULL;
+    else
+    {
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_NO_RANDR_EXTENSION,
+		     _("RANDR extension is not present"));
+	
+	return NULL;
+   }
 }
 
 void
@@ -590,10 +634,25 @@ gnome_rr_screen_get_ranges (GnomeRRScreen *screen,
 	*max_height = screen->info->max_height;
 }
 
+/**
+ * gnome_rr_screen_refresh
+ * @screen: a #GnomeRRScreen
+ * @error: location to store error, or %NULL
+ *
+ * Refreshes the screen configuration, and calls the screen's callback if it
+ * exists and if the screen's configuration changed.
+ *
+ * Return value: TRUE if the screen's configuration changed; otherwise, the
+ * function returns FALSE and a NULL error if the configuration didn't change,
+ * or FALSE and a non-NULL error if there was an error while refreshing the
+ * configuration.
+ */
 gboolean
-gnome_rr_screen_refresh (GnomeRRScreen *screen)
+gnome_rr_screen_refresh (GnomeRRScreen *screen,
+			 GError       **error)
 {
-    return screen_update (screen, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+    return screen_update (screen, FALSE, error);
 }
 
 GnomeRRMode **
@@ -735,8 +794,8 @@ read_edid_data (GnomeRROutput *output)
     return NULL;
 }
 
-static void
-output_initialize (GnomeRROutput *output, XRRScreenResources *res)
+static gboolean
+output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **error)
 {
     XRROutputInfo *info = XRRGetOutputInfo (
 	DISPLAY (output), res, output->id);
@@ -749,8 +808,11 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res)
     
     if (!info || !output->info)
     {
-	/* FIXME */
-	return;
+	/* FIXME: see the comment in crtc_initialize() */
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+		     _("could not get information about output %d"),
+		     (int) output->id);
+	return FALSE;
     }
     
     output->name = g_strdup (info->name); /* FIXME: what is nameLen used for? */
@@ -802,11 +864,18 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res)
     output->edid_data = read_edid_data (output);
     
     XRRFreeOutputInfo (info);
+
+    return TRUE;
 }
 
 static void
 output_free (GnomeRROutput *output)
 {
+    g_free (output->clones);
+    g_free (output->modes);
+    g_free (output->possible_crtcs);
+    g_free (output->edid_data);
+    g_free (output->name);
     g_free (output);
 }
 
@@ -1015,22 +1084,38 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
 			  GnomeRRMode      *mode,
 			  GnomeRRRotation   rotation,
 			  GnomeRROutput   **outputs,
-			  int               n_outputs)
+			  int               n_outputs,
+			  GError          **error)
 {
     ScreenInfo *info;
     GArray *output_ids;
+    Status status;
     gboolean result;
     int i;
     
     g_return_val_if_fail (crtc != NULL, FALSE);
     g_return_val_if_fail (mode != NULL || outputs == NULL || n_outputs == 0, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
     
     info = crtc->info;
     
     if (mode)
     {
-	g_return_val_if_fail (x + mode->width <= info->max_width, FALSE);
-	g_return_val_if_fail (y + mode->height <= info->max_height, FALSE);
+	if (x + mode->width > info->max_width
+	    || y + mode->height > info->max_height)
+	{
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
+			 /* Translators: the "position", "size", and "maximum"
+			  * words here are not keywords; please translate them
+			  * as usual.  */
+			 _("requested position/size for CRTC %d is outside the allowed limit: "
+			   "position=(%d, %d), size=(%d, %d), maximum=(%d, %d)"),
+			 (int) crtc->id,
+			 x, y,
+			 mode->width, mode->height,
+			 info->max_width, info->max_height);
+	    return FALSE;
+	}
     }
     
     output_ids = g_array_new (FALSE, FALSE, sizeof (RROutput));
@@ -1041,15 +1126,24 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
 	    g_array_append_val (output_ids, outputs[i]->id);
     }
     
-    result = XRRSetCrtcConfig (DISPLAY (crtc), info->resources, crtc->id,
+    status = XRRSetCrtcConfig (DISPLAY (crtc), info->resources, crtc->id,
 			       CurrentTime, 
 			       x, y,
 			       mode ? mode->id : None,
 			       xrotation_from_rotation (rotation),
 			       (RROutput *)output_ids->data,
-			       output_ids->len) == RRSetConfigSuccess;
+			       output_ids->len);
     
     g_array_free (output_ids, TRUE);
+
+    if (status == RRSetConfigSuccess)
+	result = TRUE;
+    else {
+	result = FALSE;
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+		     _("could not set the configuration for CRTC %d"),
+		     (int) crtc->id);
+    }
     
     return result;
 }
@@ -1137,9 +1231,10 @@ crtc_new (ScreenInfo *info, RROutput id)
     return crtc;
 }
 
-static void
+static gboolean
 crtc_initialize (GnomeRRCrtc        *crtc,
-		 XRRScreenResources *res)
+		 XRRScreenResources *res,
+		 GError            **error)
 {
     XRRCrtcInfo *info = XRRGetCrtcInfo (DISPLAY (crtc), res, crtc->id);
     GPtrArray *a;
@@ -1152,7 +1247,12 @@ crtc_initialize (GnomeRRCrtc        *crtc,
     if (!info)
     {
 	/* FIXME: We need to reaquire the screen resources */
-	return;
+	/* FIXME: can we actually catch BadRRCrtc, and does it make sense to emit that? */
+
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+		     _("could not get information about CRTC %d"),
+		     (int) crtc->id);
+	return FALSE;
     }
     
     /* GnomeRRMode */
@@ -1190,6 +1290,8 @@ crtc_initialize (GnomeRRCrtc        *crtc,
     crtc->rotations = gnome_rr_rotation_from_xrotation (info->rotations);
     
     XRRFreeCrtcInfo (info);
+
+    return TRUE;
 }
 
 static void
